# Project : PaypalLog2QIF
# Converts a TAB-delimited Paypal log file in QIF format to be imported in MsMoney


#define the format for the STDOUT
format fmtSTDOUTFORMAT_HEADER =
       Date     Amount   Payee
------------------------------------------------------------------------------
.
format fmtSTDOUTFORMAT =
@>>>>>>>>>> @######.##   @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$TransDate, $TransVal,   $TransPayee
.

# we'll use this variable to calculate the total of transactions
$lfTransactionTotal = 0.0;

# A hash to get the right months values
%mapMonths = (  "Jan", "01",
				"Feb", "02",
				"Mar", "03",
				"Apr", "04",
				"May", "05",
				"Jun", "06",
				"Jul", "07",
				"Aug", "08",
				"Sep", "09",
				"Oct", "10",
				"Nov", "11",
				"Dec", "12",
			 );

##############################################################################
# Function vWriteTransactionSummary
#
# Purpose: Write a QIF transaction line in the output file
# Arguments: Date, Value, Amount
##############################################################################
sub vWriteTransactionSummary
{
	($TransDate, $TransVal, $TransPayee) = @_;
	# display on stdout read data
	write STDOUT; 
	# update the transation total 
	$lfTransactionTotal = $lfTransactionTotal + $TransVal;
}


##############################################################################
# Function vWriteQIFTransaction
#
# Purpose: Write a QIF transaction line in the output file
# Arguments: Date, Amount, Payee, Memo
##############################################################################
sub vWriteQIFTransaction
{
	my ($strDate, $strAmount, $strPayee, $strMemo) = @_;
	die "Incorrect Arguments!" if !($strDate && $strAmount && $strPayee && $strMemo);

	# display transaction summary to stdout
	vWriteTransactionSummary($strDate,$strAmount,$strPayee);
	
	# write transaction start
	print OUTPUTFILE "!Type:Cash\n";

	#write date
	print OUTPUTFILE "D" . $strDate . "\n";

	#write amount
	print OUTPUTFILE "T" . $strAmount . "\n";
	
	#write payee
	print OUTPUTFILE "P" . $strPayee . "\n";
	
	#write memo fields
	print OUTPUTFILE "M" . $strMemo . "\n";
	
	# write transaction status: we'll write them all as cleared
	print OUTPUTFILE "CX\n";
	
	# write transaction end
	print OUTPUTFILE "^\n";
}

sub strGetField
{
	my ($fieldIndex, @rgFields) = @_;
	if (!defined($fieldIndex) || $fieldIndex < 0 || $fieldIndex >= @rgFields)
	{
		return "";
	}
	return $rgFields[$fieldIndex];
}

##############################################################################
#
# MAIN Program start
#
##############################################################################
$strInputFileName = $ARGV[0];
if (!$strInputFileName)
{
	print "Usage: PaypalLog2QIF <PayPal_Log_File>\n";
	return;
}

# open input file name
open INPUTFILE, $strInputFileName or die "Cannot open PayPal log file";

# output file would have the same name as the input file and QIF extension
$strOutputFile = $strInputFileName . ".QIF";
open OUTPUTFILE, ">" . $strOutputFile or die "Cannot open output file $strOutputFile";

# read the first line, it contains the fields description
my %mapFieldsDefs = ();
$strLine = <INPUTFILE>;
my @rgFields = split /\t/, $strLine;
for (my $i = 0; $i < @rgFields; $i++)
{ 
	$strField = $rgFields[$i];
	$strField =~ s/^\s*(.*?)\s*$/$1/;
	if ($strField)
	{
		$mapFieldsDefs{$strField} = $i;
	}
}

#select the formats for STDOUT
$^ = "fmtSTDOUTFORMAT_HEADER";
$~ = "fmtSTDOUTFORMAT";
$= = 999999;

# read until the end of file
# Since the QIF file format does not have a field for the transaction time,
# MsMoney is not able to sort correctly the transactions if we write them 
# as they are in the Paypal log (in reverse order). Therefore we'll read 
# all the lines in a list, in reverse order, then parse the list.
while ($strLine = <INPUTFILE>)
{
	# while the line does not end in TAB, read more from the file
	while (!($strLine =~ /\"[\t]$/) )
	{
		# try to read next line as part of the previous unfinished line
		if (!($strPart = <INPUTFILE>))
		{
			# something must be wrong, last line is not terminated
			die "Last line in the log file is not terminated";
		}
		
		$strLine = $strLine . $strPart;
	}
	
	@lstTransactionList = ($strLine, @lstTransactionList);
}

foreach $strLine (@lstTransactionList)  
{
	# split the line, fields are delimited by a tab between quotation marks
	@rgFields = split /\"[\t]\"/, $strLine;
	# cut the first quote from the first field
	substr($rgFields[0], 0, 1) = '';
	# and the quote and tab from the last field
	substr($rgFields[@rgFields-1], length($rgFields[@rgFields-1])-2) = '';
	
	$strDate                = strGetField($mapFieldsDefs{"Date"}, @rgFields);
	$strTime                = strGetField($mapFieldsDefs{"Time"}, @rgFields);
	$strName                = strGetField($mapFieldsDefs{"Name"}, @rgFields);
	$strType                = strGetField($mapFieldsDefs{"Type"}, @rgFields);
	$strStatus              = strGetField($mapFieldsDefs{"Status"}, @rgFields);  
	$strGross               = strGetField($mapFieldsDefs{"Gross"}, @rgFields);
	$strFee                 = strGetField($mapFieldsDefs{"Fee"}, @rgFields);
	$strNet                 = strGetField($mapFieldsDefs{"Net"}, @rgFields);
	$strFromEmail           = strGetField($mapFieldsDefs{"From Email Address"}, @rgFields);
	$strToEmail             = strGetField($mapFieldsDefs{"To Email Address"}, @rgFields);
	$strTransactionID       = strGetField($mapFieldsDefs{"Transaction ID"}, @rgFields);
	$strPaymentType         = strGetField($mapFieldsDefs{"Payment Type"}, @rgFields);
	$strItemTitle           = strGetField($mapFieldsDefs{"Item Title"}, @rgFields);
	$strItemID              = strGetField($mapFieldsDefs{"Item ID"}, @rgFields);
	$strAuctionSite         = strGetField($mapFieldsDefs{"Auction Site"}, @rgFields);
	$strClosingDate         = strGetField($mapFieldsDefs{"Closing Date"}, @rgFields);
	#Subject is no longer used in the new format
	$strSubject             = strGetField($mapFieldsDefs{"Subject"}, @rgFields);
	# In the old format, the Auction Site was written as Auction House
	if (!$strAuctionSite)
	{
		$strAuctionSite	= strGetField($mapFieldsDefs{"Auction House"}, @rgFields);
	}

	# At some time, Paypal used currently a user-friendly format for the date (like "Jan 02, 2003"), 
	# that made Money unable to import the downloaded log. 
	# If that's the case, we'll convert this date into an understandable date format (like "01/02/2003")
	if ($strDate =~ /(\w*)\s*(\d*)\,\s*(\d*)/)
	{
		$strMonth = $1;
		$strDay = $2;
		$strYear = $3;
		if (defined $mapMonths{$strMonth})
		{
			$strDate = $mapMonths{$strMonth} . '/' . $strDay . '/' . $strYear;
		}
	}
	
	# prepare the fields for the main transaction
	if ($strSubject || $strAuctionSite || $strItemTitle)
	{
		# write payee as 'Name (Email)'
		$strPayee = $strName;
		if ($strGross > 0.0 )
		{
			if ($strFromEmail)
			{
				$strPayee = $strPayee . " (" . $strFromEmail . ")";
			}
		}
		else
		{
			if ($strFromEmail)
			{
				$strPayee = $strPayee . " (" . $strToEmail . ")";
			}
		}

		# write memo as 'ItemName (AuctionSite AuctionID)'		
		$strMemo = $strSubject;
		if (!$strMemo)
		{
			$strMemo = $strItemTitle;
		}
		
		if ($strAuctionSite)
		{
			$strMemo = $strMemo . " (" . $strAuctionSite; 
			if ($strItemID)
			{
				$strMemo = $strMemo . " " . $strItemID;
			}
			
			$strMemo = $strMemo . ")";
		}
	}
	else
	{
		# write payee in detail
		$strPayee = $strType;
		
		# write memo as 'Account [xxx-xxx-xxx] TransactionID'		
		$strMemo = $strPaymentType;
		if ($strMemo && $strTransactionID)
		{
			$strMemo = $strMemo . " ";
		}
		
		$strMemo = $strMemo . $strTransactionID;
	}

	vWriteQIFTransaction $strDate, $strGross, $strPayee, $strMemo;
	
	# if the transaction had associated fees from Paypal, we'll write a second transaction
	if ($strFee != 0.0)
	{
		$strMemo = "Fees for transaction \"" . $strPayee . "\" - " . $strMemo;
		$strPayee = "PayPal fees";
		vWriteQIFTransaction $strDate, $strFee, $strPayee, $strMemo;
	}
}

# display on stdout total of transactions
print STDOUT "\n\nTotal transaction processed: $lfTransactionTotal\$\n"; 

close INPUTFILE;
close OUTPUTFILE;